@@ -1,29 +1,59 @@
 import functools
-import logging
 import json
-from os import environ, getpid
+import logging
+import os
+from os import environ
+from os import getpid
 import sys
+from typing import Any
+from typing import Callable
+from typing import Dict
+from typing import List
+from typing import Optional
+from typing import Set
+from typing import Union
 
+from ddtrace import config
+from ddtrace.filters import TraceFilter
 from ddtrace.vendor import debtcollector
 
-from .constants import FILTERS_KEY, SAMPLE_RATE_METRIC_KEY, VERSION_KEY, ENV_KEY
-from .ext import system
-from .ext.priority import AUTO_REJECT, AUTO_KEEP
-from .internal import debug
-from .internal.logger import get_logger, hasHandlers
-from .internal.runtime import RuntimeTags, RuntimeWorker, get_runtime_id
-from .internal.writer import AgentWriter, LogWriter
-from .internal import _rand
-from .provider import DefaultContextProvider
-from .context import Context
-from .sampler import DatadogSampler, RateSampler, RateByServiceSampler
-from .settings import config
-from .span import Span
-from .utils.formats import asbool, get_env
-from .utils.deprecation import deprecated, RemovedInDDTrace10Warning
-from .vendor.dogstatsd import DogStatsd
-from . import compat
 from . import _hooks
+from .constants import ENV_KEY
+from .constants import FILTERS_KEY
+from .constants import HOSTNAME_KEY
+from .constants import SAMPLE_RATE_METRIC_KEY
+from .constants import VERSION_KEY
+from .context import Context
+from .ext import system
+from .ext.priority import AUTO_KEEP
+from .ext.priority import AUTO_REJECT
+from .internal import _rand
+from .internal import agent
+from .internal import atexit
+from .internal import compat
+from .internal import debug
+from .internal import hostname
+from .internal.dogstatsd import get_dogstatsd_client
+from .internal.logger import get_logger
+from .internal.logger import hasHandlers
+from .internal.processor import SpanProcessor
+from .internal.processor.trace import SpanAggregator
+from .internal.processor.trace import TraceProcessor
+from .internal.processor.trace import TraceSamplingProcessor
+from .internal.runtime import get_runtime_id
+from .internal.writer import AgentWriter
+from .internal.writer import LogWriter
+from .internal.writer import TraceWriter
+from .provider import DefaultContextProvider
+from .sampler import BasePrioritySampler
+from .sampler import BaseSampler
+from .sampler import DatadogSampler
+from .sampler import RateByServiceSampler
+from .sampler import RateSampler
+from .span import Span
+from .utils.deprecation import deprecated
+from .utils.formats import asbool
+from .utils.formats import get_env
 
 
 log = get_logger(__name__)
@@ -41,33 +71,7 @@ if debug_mode and not hasHandlers(log):
         logging.basicConfig(level=logging.DEBUG)
 
 
-def _parse_dogstatsd_url(url):
-    if url is None:
-        return
-
-    # url can be either of the form `udp://<host>:<port>` or `unix://<path>`
-    # also support without url scheme included
-    if url.startswith('/'):
-        url = 'unix://' + url
-    elif '://' not in url:
-        url = 'udp://' + url
-
-    parsed = compat.parse.urlparse(url)
-
-    if parsed.scheme == 'unix':
-        return dict(socket_path=parsed.path)
-    elif parsed.scheme == 'udp':
-        return dict(host=parsed.hostname, port=parsed.port)
-    else:
-        raise ValueError('Unknown scheme `%s` for DogStatsD URL `{}`'.format(parsed.scheme))
-
-
-_INTERNAL_APPLICATION_SPAN_TYPES = [
-    "custom",
-    "template",
-    "web",
-    "worker"
-]
+_INTERNAL_APPLICATION_SPAN_TYPES = {"custom", "template", "web", "worker"}
 
 
 class Tracer(object):
@@ -81,84 +85,76 @@ class Tracer(object):
         from ddtrace import tracer
         trace = tracer.trace('app.request', 'web-server').finish()
     """
-    _RUNTIME_METRICS_INTERVAL = 10
 
-    DEFAULT_HOSTNAME = environ.get('DD_AGENT_HOST', environ.get('DATADOG_TRACE_AGENT_HOSTNAME', 'localhost'))
-    DEFAULT_PORT = int(environ.get('DD_TRACE_AGENT_PORT', 8126))
-    DEFAULT_DOGSTATSD_PORT = int(get_env('dogstatsd', 'port', default=8125))
-    DEFAULT_DOGSTATSD_URL = get_env('dogstatsd', 'url',
-                                    default='udp://{}:{}'.format(DEFAULT_HOSTNAME, DEFAULT_DOGSTATSD_PORT))
-    DEFAULT_AGENT_URL = environ.get('DD_TRACE_AGENT_URL', 'http://%s:%d' % (DEFAULT_HOSTNAME, DEFAULT_PORT))
+    SHUTDOWN_TIMEOUT = 5
 
-    def __init__(self, url=None, dogstatsd_url=DEFAULT_DOGSTATSD_URL):
+    def __init__(
+        self,
+        url=None,  # type: Optional[str]
+        dogstatsd_url=None,  # type: Optional[str]
+    ):
+        # type: (...) -> None
         """
         Create a new ``Tracer`` instance. A global tracer is already initialized
         for common usage, so there is no need to initialize your own ``Tracer``.
 
         :param url: The Datadog agent URL.
-        :param url: The DogStatsD URL.
+        :param dogstatsd_url: The DogStatsD URL.
         """
         self.log = log
-        self.sampler = None
-        self.priority_sampler = None
-        self._runtime_worker = None
-
-        uds_path = None
-        https = None
-        hostname = self.DEFAULT_HOSTNAME
-        port = self.DEFAULT_PORT
-        writer = None
-
-        if self._is_agentless_environment() and url is None:
-            writer = LogWriter()
-        else:
-            if url is None:
-                url = self.DEFAULT_AGENT_URL
-            url_parsed = compat.parse.urlparse(url)
-            if url_parsed.scheme in ('http', 'https'):
-                hostname = url_parsed.hostname
-                port = url_parsed.port
-                https = url_parsed.scheme == 'https'
-                # FIXME This is needed because of the way of configure() works right now, where it considers `port=None`
-                # to be "no port set so let's use the default".
-                # It should go away when we remove configure()
-                if port is None:
-                    if https:
-                        port = 443
-                    else:
-                        port = 80
-            elif url_parsed.scheme == 'unix':
-                uds_path = url_parsed.path
-            else:
-                raise ValueError('Unknown scheme `%s` for agent URL' % url_parsed.scheme)
+        self._filters = []  # type: List[TraceFilter]
 
         # globally set tags
         self.tags = config.tags.copy()
 
         # a buffer for service info so we don't perpetually send the same things
-        self._services = set()
+        self._services = set()  # type: Set[str]
 
         # Runtime id used for associating data collected during runtime to
         # traces
         self._pid = getpid()
 
         self.enabled = asbool(get_env("trace", "enabled", default=True))
+        self.context_provider = DefaultContextProvider()
+        self.sampler = DatadogSampler()  # type: BaseSampler
+        self.priority_sampler = RateByServiceSampler()  # type: Optional[BasePrioritySampler]
+        self._dogstatsd_url = agent.get_stats_url() if dogstatsd_url is None else dogstatsd_url
 
-        # Apply the default configuration
-        self.configure(
-            hostname=hostname,
-            port=port,
-            https=https,
-            uds_path=uds_path,
-            sampler=DatadogSampler(),
-            context_provider=DefaultContextProvider(),
-            dogstatsd_url=dogstatsd_url,
-            writer=writer,
+        if self._use_log_writer() and url is None:
+            writer = LogWriter()  # type: TraceWriter
+        else:
+            url = url or agent.get_trace_url()
+            agent.verify_url(url)
+
+            writer = AgentWriter(
+                agent_url=url,
+                sampler=self.sampler,
+                priority_sampler=self.priority_sampler,
+                dogstatsd=get_dogstatsd_client(self._dogstatsd_url),
+                report_metrics=config.health_metrics_enabled,
+                sync_mode=self._use_sync_mode(),
+            )
+        self.writer = writer  # type: TraceWriter
+        self._partial_flush_enabled = asbool(get_env("tracer", "partial_flush_enabled", default=False))
+        self._partial_flush_min_spans = int(
+            get_env("tracer", "partial_flush_min_spans", default=500)  # type: ignore[arg-type]
         )
-
+        self._initialize_span_processors()
         self._hooks = _hooks.Hooks()
+        atexit.register(self._atexit)
+
+    def _atexit(self):
+        # type: () -> None
+        key = "ctrl-break" if os.name == "nt" else "ctrl-c"
+        log.debug(
+            "Waiting %d seconds for tracer to finish. Hit %s to quit.",
+            self.SHUTDOWN_TIMEOUT,
+            key,
+        )
+        self.shutdown(timeout=self.SHUTDOWN_TIMEOUT)
 
     def on_start_span(self, func):
+        # type: (Callable) -> Callable
         """Register a function to execute when a span start.
 
         Can be used as a decorator.
@@ -170,6 +166,7 @@ class Tracer(object):
         return func
 
     def deregister_on_start_span(self, func):
+        # type: (Callable) -> Callable
         """Unregister a function registered to execute when a span starts.
 
         Can be used as a decorator.
@@ -184,20 +181,22 @@ class Tracer(object):
     def debug_logging(self):
         return self.log.isEnabledFor(logging.DEBUG)
 
-    @debug_logging.setter
-    @deprecated(message='Use logging.setLevel instead', version='1.0.0')
+    @debug_logging.setter  # type: ignore[misc]
+    @deprecated(message="Use logging.setLevel instead", version="1.0.0")
     def debug_logging(self, value):
+        # type: (bool) -> None
         self.log.setLevel(logging.DEBUG if value else logging.WARN)
 
-    @deprecated('Use .tracer, not .tracer()', '1.0.0')
+    @deprecated("Use .tracer, not .tracer()", "1.0.0")
     def __call__(self):
         return self
 
-    @deprecated('This method will be removed altogether', '1.0.0')
+    @deprecated("This method will be removed altogether", "1.0.0")
     def global_excepthook(self, tp, value, traceback):
         """The global tracer except hook."""
 
     def get_call_context(self, *args, **kwargs):
+        # type: (...) -> Context
         """
         Return the current active ``Context`` for this traced execution. This method is
         automatically called in the ``tracer.trace()``, but it can be used in the application
@@ -213,31 +212,29 @@ class Tracer(object):
         This method makes use of a ``ContextProvider`` that is automatically set during the tracer
         initialization, or while using a library instrumentation.
         """
-        return self.context_provider.active(*args, **kwargs)
+        return self.context_provider.active(*args, **kwargs)  # type: ignore
 
     # TODO: deprecate this method and make sure users create a new tracer if they need different parameters
-    @debtcollector.removals.removed_kwarg("dogstatsd_host", "Use `dogstatsd_url` instead",
-                                          category=RemovedInDDTrace10Warning)
-    @debtcollector.removals.removed_kwarg("dogstatsd_port", "Use `dogstatsd_url` instead",
-                                          category=RemovedInDDTrace10Warning)
+    @debtcollector.removals.removed_kwarg("collect_metrics", removal_version="0.51")
     def configure(
         self,
-        enabled=None,
-        hostname=None,
-        port=None,
-        uds_path=None,
-        https=None,
-        sampler=None,
-        context_provider=None,
-        wrap_executor=None,
-        priority_sampling=None,
-        settings=None,
-        collect_metrics=None,
-        dogstatsd_host=None,
-        dogstatsd_port=None,
-        dogstatsd_url=None,
-        writer=None,
+        enabled=None,  # type: Optional[bool]
+        hostname=None,  # type: Optional[str]
+        port=None,  # type: Optional[int]
+        uds_path=None,  # type: Optional[str]
+        https=None,  # type: Optional[bool]
+        sampler=None,  # type: Optional[BaseSampler]
+        context_provider=None,  # type: Optional[DefaultContextProvider]
+        wrap_executor=None,  # type: Optional[Callable]
+        priority_sampling=None,  # type: Optional[bool]
+        settings=None,  # type: Optional[Dict[str, Any]]
+        collect_metrics=None,  # type: Optional[bool]
+        dogstatsd_url=None,  # type: Optional[str]
+        writer=None,  # type: Optional[TraceWriter]
+        partial_flush_enabled=None,  # type: Optional[bool]
+        partial_flush_min_spans=None,  # type: Optional[int]
     ):
+        # type: (...) -> None
         """
         Configure an existing Tracer the easy way.
         Allow to configure or reconfigure a Tracer instance.
@@ -258,16 +255,21 @@ class Tracer(object):
         :param priority_sampling: enable priority sampling, this is required for
             complete distributed tracing support. Enabled by default.
         :param collect_metrics: Whether to enable runtime metrics collection.
-        :param str dogstatsd_host: Host for UDP connection to DogStatsD (deprecated: use dogstatsd_url)
-        :param int dogstatsd_port: Port for UDP connection to DogStatsD (deprecated: use dogstatsd_url)
         :param str dogstatsd_url: URL for UDP or Unix socket connection to DogStatsD
         """
         if enabled is not None:
             self.enabled = enabled
 
-        filters = None
         if settings is not None:
             filters = settings.get(FILTERS_KEY)
+            if filters is not None:
+                self._filters = filters
+
+        if partial_flush_enabled is not None:
+            self._partial_flush_enabled = partial_flush_enabled
+
+        if partial_flush_min_spans is not None:
+            self._partial_flush_min_spans = partial_flush_min_spans
 
         # If priority sampling is not set or is True and no priority sampler is set yet
         if priority_sampling in (None, True) and not self.priority_sampler:
@@ -279,52 +281,58 @@ class Tracer(object):
         if sampler is not None:
             self.sampler = sampler
 
-        if dogstatsd_host is not None and dogstatsd_url is None:
-            dogstatsd_url = 'udp://{}:{}'.format(dogstatsd_host, dogstatsd_port or self.DEFAULT_DOGSTATSD_PORT)
+        self._dogstatsd_url = dogstatsd_url or self._dogstatsd_url
 
-        if dogstatsd_url is not None:
-            dogstatsd_kwargs = _parse_dogstatsd_url(dogstatsd_url)
-            self.log.debug('Connecting to DogStatsd(%s)', dogstatsd_url)
-            self._dogstatsd_client = DogStatsd(**dogstatsd_kwargs)
-
-        if writer:
-            self.writer = writer
-        elif (
-            hostname is not None
-            or port is not None
-            or uds_path is not None
-            or https is not None
-            or filters is not None
-            or priority_sampling is not None
-            or sampler is not None
-        ):
-            # Preserve hostname and port when overriding filters or priority sampling
-            # This is clumsy and a good reason to get rid of this configure() API
-            if hasattr(self, 'writer') and hasattr(self.writer, 'api'):
-                default_hostname = self.writer.api.hostname
-                default_port = self.writer.api.port
-                if https is None:
-                    https = self.writer.api.https
+        if any(x is not None for x in [hostname, port, uds_path, https]):
+            # If any of the parts of the URL have updated, merge them with
+            # the previous writer values.
+            if isinstance(self.writer, AgentWriter):
+                prev_url_parsed = compat.parse.urlparse(self.writer.agent_url)
             else:
-                default_hostname = self.DEFAULT_HOSTNAME
-                default_port = self.DEFAULT_PORT
+                prev_url_parsed = compat.parse.urlparse("")
 
-            if hasattr(self, "writer") and self.writer.is_alive():
-                self.writer.stop()
+            if uds_path is not None:
+                if hostname is None and prev_url_parsed.scheme == "unix":
+                    hostname = prev_url_parsed.hostname
+                url = "unix://%s%s" % (hostname or "", uds_path)
+            else:
+                if https is None:
+                    https = prev_url_parsed.scheme == "https"
+                if hostname is None:
+                    hostname = prev_url_parsed.hostname or ""
+                if port is None:
+                    port = prev_url_parsed.port
+                scheme = "https" if https else "http"
+                url = "%s://%s:%s" % (scheme, hostname, port)
+        elif isinstance(self.writer, AgentWriter):
+            # Reuse the URL from the previous writer if there was one.
+            url = self.writer.agent_url
+        else:
+            # No URL parts have updated and there's no previous writer to
+            # get the URL from.
+            url = None  # type: ignore
 
+        self.writer.stop()
+
+        if writer is not None:
+            self.writer = writer
+        elif url:
+            # Verify the URL and create a new AgentWriter with it.
+            agent.verify_url(url)
             self.writer = AgentWriter(
-                hostname or default_hostname,
-                port or default_port,
-                uds_path=uds_path,
-                https=https,
-                filters=filters,
+                url,
                 sampler=self.sampler,
                 priority_sampler=self.priority_sampler,
-                dogstatsd=self._dogstatsd_client,
+                dogstatsd=get_dogstatsd_client(self._dogstatsd_url),
+                report_metrics=config.health_metrics_enabled,
+                sync_mode=self._use_sync_mode(),
             )
-
-        # HACK: since we recreated our dogstatsd agent, replace the old write one
-        self.writer.dogstatsd = self._dogstatsd_client
+        elif writer is None and isinstance(self.writer, LogWriter):
+            # No need to do anything for the LogWriter.
+            pass
+        if isinstance(self.writer, AgentWriter):
+            self.writer.dogstatsd = get_dogstatsd_client(self._dogstatsd_url)
+        self._initialize_span_processors()
 
         if context_provider is not None:
             self.context_provider = context_provider
@@ -332,19 +340,19 @@ class Tracer(object):
         if wrap_executor is not None:
             self._wrap_executor = wrap_executor
 
-        # Since we've recreated our dogstatsd agent, we need to restart metric collection with that new agent
-        if self._runtime_worker:
+        runtime_metrics_was_running = False
+        # FIXME: Import RuntimeWorker here to avoid circular imports. This will
+        # be gone together with the collect_metrics attribute soon.
+        from .internal.runtime.runtime_metrics import RuntimeWorker
+
+        if RuntimeWorker._instance is not None:
             runtime_metrics_was_running = True
-            self._runtime_worker.stop()
-            self._runtime_worker.join()
-            self._runtime_worker = None
-        else:
-            runtime_metrics_was_running = False
+            RuntimeWorker.disable()
 
         if (collect_metrics is None and runtime_metrics_was_running) or collect_metrics:
-            self._start_runtime_worker()
+            RuntimeWorker.enable(tracer=self, dogstatsd_url=self._dogstatsd_url)
 
-        if debug_mode or asbool(environ.get("DD_TRACE_STARTUP_LOGS", True)):
+        if debug_mode or asbool(environ.get("DD_TRACE_STARTUP_LOGS", False)):
             try:
                 info = debug.collect(self)
             except Exception as e:
@@ -355,12 +363,22 @@ class Tracer(object):
                     msg = "- DATADOG TRACER CONFIGURATION - %s" % json.dumps(info)
                     self._log_compat(logging.INFO, msg)
 
+                # Always log errors since we're either in debug_mode or start up logs
+                # are enabled.
                 agent_error = info.get("agent_error")
                 if agent_error:
                     msg = "- DATADOG TRACER DIAGNOSTIC - %s" % agent_error
                     self._log_compat(logging.WARNING, msg)
 
-    def start_span(self, name, child_of=None, service=None, resource=None, span_type=None):
+    def start_span(
+        self,
+        name,  # type: str
+        child_of=None,  # type: Optional[Union[Span, Context]]
+        service=None,  # type: Optional[str]
+        resource=None,  # type: Optional[str]
+        span_type=None,  # type: Optional[str]
+    ):
+        # type: (...) -> Span
         """
         Return a span that will trace an operation called `name`. This method allows
         parenting using the ``child_of`` kwarg. If it's missing, the newly created span is a
@@ -391,7 +409,7 @@ class Tracer(object):
         if child_of is not None:
             if isinstance(child_of, Context):
                 context = new_ctx or child_of
-                parent = child_of.get_current_span()
+                parent = child_of._get_current_span()
             else:
                 context = child_of.context
                 parent = child_of
@@ -418,6 +436,8 @@ class Tracer(object):
             else:
                 service = config.service
 
+        mapped_service = config.service_mapping.get(service, service)
+
         if trace_id:
             # child_of a non-empty context, so either a local child span or from a remote context
             span = Span(
@@ -425,10 +445,11 @@ class Tracer(object):
                 name,
                 trace_id=trace_id,
                 parent_id=parent_span_id,
-                service=service,
+                service=mapped_service,
                 resource=resource,
                 span_type=span_type,
                 _check_pid=False,
+                on_finish=[self._on_span_finish],
             )
 
             # Extra attributes when from a local parent
@@ -441,12 +462,16 @@ class Tracer(object):
             span = Span(
                 self,
                 name,
-                service=service,
+                service=mapped_service,
                 resource=resource,
                 span_type=span_type,
                 _check_pid=False,
+                on_finish=[self._on_span_finish],
             )
-
+            span.metrics[system.PID] = self._pid or getpid()
+            span.meta["runtime-id"] = get_runtime_id()
+            if config.report_hostname:
+                span.meta[HOSTNAME_KEY] = hostname.get_hostname()
             span.sampled = self.sampler.sample(span)
             # Old behavior
             # DEV: The new sampler sets metrics and priority sampling on the span for us
@@ -474,17 +499,12 @@ class Tracer(object):
                 # We must always mark the span as sampled so it is forwarded to the agent
                 span.sampled = True
 
-            # add tags to root span to correlate trace with runtime metrics
-            # only applied to spans with types that are internal to applications
-            if self._runtime_worker and self._is_span_internal(span):
-                span.set_tag('language', 'python')
-
         # Apply default global tags.
         if self.tags:
             span.set_tags(self.tags)
 
         if config.env:
-            span.set_tag(ENV_KEY, config.env)
+            span._set_str_tag(ENV_KEY, config.env)
 
         # Only set the version tag on internal spans.
         if config.version:
@@ -493,47 +513,52 @@ class Tracer(object):
             #     2. the span is not the root, but the root span's service matches the span's service
             #        and the root span has a version tag
             # then the span belongs to the user application and so set the version tag
-            if (root_span is None and service == config.service) or \
-               (root_span and root_span.service == service and VERSION_KEY in root_span.meta):
-                span.set_tag(VERSION_KEY, config.version)
-
-        if not span._parent:
-            span.set_tag(system.PID, getpid())
-            span.set_tag("runtime-id", get_runtime_id())
+            if (root_span is None and service == config.service) or (
+                root_span and root_span.service == service and VERSION_KEY in root_span.meta
+            ):
+                span._set_str_tag(VERSION_KEY, config.version)
 
         # add it to the current context
-        context.add_span(span)
+        context._add_span(span)
 
         # update set of services handled by tracer
         if service and service not in self._services and self._is_span_internal(span):
             self._services.add(service)
 
-            # The constant tags for the dogstatsd client needs to updated with any new
-            # service(s) that may have been added.
-            self._update_dogstatsd_constant_tags()
+        for p in self._span_processors:
+            p.on_span_start(span)
 
         self._hooks.emit(self.__class__.start_span, span)
-
         return span
 
-    def _update_dogstatsd_constant_tags(self):
-        """ Prepare runtime tags for ddstatsd.
-        """
-        # DEV: ddstatsd expects tags in the form ['key1:value1', 'key2:value2', ...]
-        tags = [
-            '{}:{}'.format(k, v)
-            for k, v in RuntimeTags()
-        ]
-        self.log.debug('Updating constant tags %s', tags)
-        self._dogstatsd_client.constant_tags = tags
+    def _on_span_finish(self, span):
+        if not self.enabled:
+            return  # nothing to do
 
-    def _start_runtime_worker(self):
-        self._runtime_worker = RuntimeWorker(self._dogstatsd_client, self._RUNTIME_METRICS_INTERVAL)
-        self._runtime_worker.start()
+        if self.log.isEnabledFor(logging.DEBUG):
+            self.log.debug("finishing span %s (enabled:%s)", span.pprint(), self.enabled)
+
+        for p in self._span_processors:
+            p.on_span_finish(span)
+
+    def _initialize_span_processors(self):
+        # type: () -> None
+        trace_processors = []  # type: List[TraceProcessor]
+        trace_processors += [TraceSamplingProcessor()]
+        trace_processors += self._filters
+
+        self._span_processors = [
+            SpanAggregator(
+                partial_flush_enabled=self._partial_flush_enabled,
+                partial_flush_min_spans=self._partial_flush_min_spans,
+                trace_processors=trace_processors,
+                writer=self.writer,
+            ),
+        ]  # type: List[SpanProcessor]
 
     def _check_new_process(self):
-        """ Checks if the tracer is in a new process (was forked) and performs
-            the necessary updates if it is a new process
+        """Checks if the tracer is in a new process (was forked) and performs
+        the necessary updates if it is a new process
         """
         pid = getpid()
         if self._pid == pid:
@@ -562,16 +587,9 @@ class Tracer(object):
         # of the parent.
         self._services = set()
 
-        if self._runtime_worker is not None:
-            self._start_runtime_worker()
-
-        # force an immediate update constant tags since we have reset services
-        # and generated a new runtime id
-        self._update_dogstatsd_constant_tags()
-
         # Re-create the background writer thread
         self.writer = self.writer.recreate()
-
+        self._initialize_span_processors()
         return new_ctx
 
     def _log_compat(self, level, msg):
@@ -592,6 +610,7 @@ class Tracer(object):
             self.log.log(level, msg)
 
     def trace(self, name, service=None, resource=None, span_type=None):
+        # type: (str, Optional[str], Optional[str], Optional[str]) -> Span
         """
         Return a span that will trace an operation called `name`. The context that created
         the span as well as the span parenting, are automatically handled by the tracing
@@ -639,6 +658,7 @@ class Tracer(object):
         )
 
     def current_root_span(self):
+        # type: () -> Optional[Span]
         """Returns the root span of the current context.
 
         This is useful for attaching information related to the trace as a
@@ -654,29 +674,22 @@ class Tracer(object):
         """
         ctx = self.get_call_context()
         if ctx:
-            return ctx.get_current_root_span()
+            return ctx._get_current_root_span()
         return None
 
     def current_span(self):
+        # type: () -> Optional[Span]
         """
         Return the active span for the current call context or ``None``
         if no spans are available.
         """
         ctx = self.get_call_context()
         if ctx:
-            return ctx.get_current_span()
+            return ctx._get_current_span()
         return None
 
-    def record(self, context):
-        """
-        Record the given ``Context`` if it's finished.
-        """
-        # extract and enqueue the trace if it's sampled
-        trace, sampled = context.get()
-        if trace and sampled:
-            self.write(trace)
-
     def write(self, spans):
+        # type: (Optional[List[Span]]) -> None
         """
         Send the trace to the writer to enqueue the spans list in the agent
         sending queue.
@@ -685,21 +698,29 @@ class Tracer(object):
             return  # nothing to do
 
         if self.log.isEnabledFor(logging.DEBUG):
-            self.log.debug('writing %s spans (enabled:%s)', len(spans), self.enabled)
+            self.log.debug("writing %s spans (enabled:%s)", len(spans), self.enabled)
             for span in spans:
-                self.log.debug('\n%s', span.pprint())
+                self.log.debug("\n%s", span.pprint())
 
-        if self.enabled and self.writer:
-            # only submit the spans if we're actually enabled (and don't crash :)
+        if not self.enabled:
+            return
+
+        if spans is not None:
             self.writer.write(spans=spans)
 
-    @deprecated(message='Manually setting service info is no longer necessary', version='1.0.0')
+    @deprecated(message="Manually setting service info is no longer necessary", version="1.0.0")
     def set_service_info(self, *args, **kwargs):
-        """Set the information about the given service.
-        """
+        """Set the information about the given service."""
         return
 
-    def wrap(self, name=None, service=None, resource=None, span_type=None):
+    def wrap(
+        self,
+        name=None,  # type: Optional[str]
+        service=None,  # type: Optional[str]
+        resource=None,  # type: Optional[str]
+        span_type=None,  # type: Optional[str]
+    ):
+        # type: (...) -> Callable[[Callable[..., Any]], Callable[..., Any]]
         """
         A decorator used to trace an entire function. If the traced function
         is a coroutine, it traces the coroutine execution when is awaited.
@@ -741,9 +762,10 @@ class Tracer(object):
                 span = tracer.current_span()
                 span.set_tag('a', 'b')
         """
+
         def wrap_decorator(f):
             # FIXME[matt] include the class name for methods.
-            span_name = name if name else '%s.%s' % (f.__module__, f.__name__)
+            span_name = name if name else "%s.%s" % (f.__module__, f.__name__)
 
             # detect if the the given function is a coroutine to use the
             # right decorator; this initial check ensures that the
@@ -754,20 +776,25 @@ class Tracer(object):
                 # code is used for compatibility reasons to prevent Syntax errors
                 # in Python 2
                 func_wrapper = compat.make_async_decorator(
-                    self, f, span_name,
+                    self,
+                    f,
+                    span_name,
                     service=service,
                     resource=resource,
                     span_type=span_type,
                 )
             else:
+
                 @functools.wraps(f)
                 def func_wrapper(*args, **kwargs):
                     # if a wrap executor has been configured, it is used instead
                     # of the default tracing function
-                    if getattr(self, '_wrap_executor', None):
+                    if getattr(self, "_wrap_executor", None):
                         return self._wrap_executor(
                             self,
-                            f, args, kwargs,
+                            f,
+                            args,
+                            kwargs,
                             span_name,
                             service=service,
                             resource=resource,
@@ -779,10 +806,12 @@ class Tracer(object):
                         return f(*args, **kwargs)
 
             return func_wrapper
+
         return wrap_decorator
 
     def set_tags(self, tags):
-        """ Set some tags at the tracer level.
+        # type: (Dict[str, str]) -> None
+        """Set some tags at the tracer level.
         This will append those tags to each span created by the tracer.
 
         :param dict tags: dict of tags to set at tracer level
@@ -790,6 +819,7 @@ class Tracer(object):
         self.tags.update(tags)
 
     def shutdown(self, timeout=None):
+        # type: (Optional[float]) -> None
         """Shutdown the tracer.
 
         This will stop the background writer/worker and flush any finished traces in the buffer.
@@ -798,24 +828,62 @@ class Tracer(object):
             before exiting or :obj:`None` to block until flushing has successfully completed (default: :obj:`None`)
         :type timeout: :obj:`int` | :obj:`float` | :obj:`None`
         """
-        if not self.writer.is_alive():
-            return
-
-        self.writer.stop()
-        self.writer.join(timeout=timeout)
+        self.writer.stop(timeout=timeout)
+        atexit.unregister(self._atexit)
 
     @staticmethod
-    def _is_agentless_environment():
-        if environ.get('DD_AGENT_HOST') or \
-           environ.get('DATADOG_TRACE_AGENT_HOSTNAME') or \
-           environ.get('DD_TRACE_AGENT_URL'):
+    def _use_log_writer():
+        # type: () -> bool
+        """Returns whether the LogWriter should be used in the environment by
+        default.
+
+        The LogWriter required by default in AWS Lambdas when the Datadog Agent extension
+        is not available in the Lambda.
+        """
+        if (
+            environ.get("DD_AGENT_HOST")
+            or environ.get("DATADOG_TRACE_AGENT_HOSTNAME")
+            or environ.get("DD_TRACE_AGENT_URL")
+        ):
             # If one of these variables are set, we definitely have an agent
             return False
-        if environ.get("AWS_LAMBDA_FUNCTION_NAME"):
-            # We are in an AWS Lambda environment
-            return True
-        return False
+        elif _in_aws_lambda() and _has_aws_lambda_agent_extension():
+            # If the Agent Lambda extension is available then an AgentWriter is used.
+            return False
+        else:
+            return _in_aws_lambda()
+
+    @staticmethod
+    def _use_sync_mode():
+        # type: () -> bool
+        """Returns, if an `AgentWriter` is to be used, whether it should be run
+         in synchronous mode by default.
+
+        There is only one case in which this is desirable:
+
+        - AWS Lambdas can have the Datadog agent installed via an extension.
+          When it's available traces must be sent synchronously to ensure all
+          are received before the Lambda terminates.
+        """
+        return _in_aws_lambda() and _has_aws_lambda_agent_extension()
 
     @staticmethod
     def _is_span_internal(span):
         return not span.span_type or span.span_type in _INTERNAL_APPLICATION_SPAN_TYPES
+
+
+def _has_aws_lambda_agent_extension():
+    # type: () -> bool
+    """Returns whether the environment has the AWS Lambda Datadog Agent
+    extension available.
+    """
+    return os.path.exists("/opt/extensions/datadog-agent")
+
+
+def _in_aws_lambda():
+    # type: () -> bool
+    """Returns whether the environment is an AWS Lambda.
+    This is accomplished by checking if the AWS_LAMBDA_FUNCTION_NAME environment
+    variable is defined.
+    """
+    return bool(environ.get("AWS_LAMBDA_FUNCTION_NAME", False))
